/***************************************************************************
    DbxReader - DBX-handling routines

    author:  Ulrich Krebs
    email :  ukrebs@freenet.de
    credits: Part of this software was taken from
             the libdbx package from David Smith
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream.h>
#include <afx.h>
#include "DbxReader.h"

void CDbxReader::Init()
{
	receiveDates=NULL;
	sendDates=NULL;
	int signature[4];
	for(int i=0;i<4;i++)
	{
		f->Read(&(signature[i]),4);
	}
	if((signature[0]==0xFE12ADCF) && (signature[1]==0x6F74FDC5) && 
		(signature[2]==0x11D1E366) && (signature[3]==0xC0004E9A))
	{
		dbxType=TYPE_EMAIL;
	}
  else if((signature[0]==0x36464D4A) && (signature[1]==0x00010003)) /* OE4 SIGNATURE */
	{
		dbxType=TYPE_OE4;
	}
	else if((signature[0]==0xFE12ADCF) && (signature[1]==0x6F74FDC6) && 
		(signature[2]==0x11D1E366) && (signature[3]==0xC0004E9A))
	{
		dbxType=TYPE_FOLDER;
	}
	else
	{
		dbxType=TYPE_UNKNOWN;
	}

	if(dbxType==TYPE_EMAIL)
	{
		ReadIndexes();
		ReadInfos();
	}
}

CDbxReader::CDbxReader(CFile *f)
{
	internalFile=false;
	this->f=f;
	Init();
} 

CDbxReader::CDbxReader(CString fileName)
{
	internalFile=true;
	f=new CFile(fileName,CFile::modeRead);
	Init();
}

CDbxReader::~CDbxReader()
{
	if(f!=NULL && internalFile)
	{
		f->Close();
		delete(f);
	}
	delete(receiveDates);
	delete(sendDates);
}

int CDbxReader::GetType()
{
	return dbxType;
}

int CDbxReader::GetItemCount()
{
	return indexes.GetSize();
}

int CDbxReader::GetIndex(int i)
{
	return indexes.GetAt(i);
}

CString CDbxReader::GetMessage(int msgNumber)
{
	DWORD index=GetIndex(msgNumber);

	DWORD size;
	BYTE count=0;

	f->Seek(index+4,CFile::begin);
	f->Read(&size,4);
	f->Seek(2,CFile::current);
	f->Read(&count,1);
	f->Seek(1,CFile::current);
	
	DWORD msgOffset=0;
	DWORD msgOffsetPtr=0;
	DWORD value=0;
	BYTE type=0;

	DWORD i;

	for(i=0;i<count;i++)
	{
		type=0;
		f->Read(&type,1);
		value=0;
		f->Read(&value,3);
		
		if(type==0x84)
		{
			msgOffset=value;
			break;
		}
		if(type==0x04)
		{
			msgOffsetPtr=index+12+value+4*count;
			break;
		}
	}
	
	DWORD totalSize=0;
	WORD blockSize=0;

	if(msgOffset==0 && msgOffsetPtr!=0)
	{
		f->Seek(msgOffsetPtr,CFile::begin);
		f->Read(&msgOffset,4);
	}

	i=msgOffset;
	DWORD oldIndex;
	while(i!=0)
	{
		if(i>=f->GetLength())
		{
			AfxThrowFileException(CFileException::badSeek);
		}

		oldIndex=i;

		f->Seek(i+8,CFile::begin);
		blockSize=0;
		f->Read(&blockSize,2);
		f->Seek(2,CFile::current);
		f->Read(&i,4);

		if(i==oldIndex)
		{
			AfxThrowFileException(CFileException::generic);
		}

		totalSize+=blockSize;
	}

	char *buf=new char[totalSize+1];

	DWORD pos=0;

	i=msgOffset;
	while(i!=0)
	{
		f->Seek(i+8,CFile::begin);
		blockSize=0;
		f->Read(&blockSize,2);
		f->Seek(2,CFile::current);
		f->Read(&i,4);

		f->Read(&buf[pos],blockSize);
		pos+=blockSize;
	}

	buf[totalSize]='\0';
	CString message(buf);
	delete(buf);

	return message;
}

CString CDbxReader::GetFileName()
{
	return f->GetFileName();
}

CTime CDbxReader::GetFileDate()
{
	CFileStatus fileStatus;
	f->GetStatus(fileStatus);
	return fileStatus.m_mtime;
}

CString CDbxReader::GetSender(int msgNumber)
{
	return senders.GetAt(msgNumber);
}

CString CDbxReader::GetSenderAddress(int msgNumber)
{
	return senderAddresses.GetAt(msgNumber);
}

CString CDbxReader::GetReceiver(int msgNumber)
{
	return receivers.GetAt(msgNumber);
}

CString CDbxReader::GetReceiverAddress(int msgNumber)
{
	return receiverAddresses.GetAt(msgNumber);
}

CString CDbxReader::GetSubject(int msgNumber)
{
	return subjects.GetAt(msgNumber);
}

CTime CDbxReader::GetReceiveDate(int msgNumber)
{
	return receiveDates[msgNumber];
}

CTime CDbxReader::GetSendDate(int msgNumber)
{
	return sendDates[msgNumber];
}

void CDbxReader::ReadInfos()
{
	receiveDates=new CTime[GetItemCount()];
	sendDates=new CTime[GetItemCount()];
	
	for(int i=0;i<GetItemCount();i++)
	{
		DWORD index=GetIndex(i);
		f->Seek(index+4,CFile::begin);
		DWORD size;
		f->Read(&size,4);
		f->Seek(2,CFile::current);
		int count=0;
		f->Read(&count,1);
		f->Seek(1,CFile::current);
		
		DWORD offset=0;
		CTime t;

		bool sender=false;
		bool senderAddress=false;
		bool receiver=false;
		bool receiverAddress=false;
		bool subject=false;
		bool receiveDate=false;
		bool sendDate=false;

		DWORD pos=index+12;
		
		for(int j=0;j<count;j++)
		{
			f->Seek(pos,CFile::begin);
			int type=0;
			f->Read(&type,1);
			int value=0;
			f->Read(&value,3);
			
			offset=index+12+4*count+value;
			switch(type)
			{
				case 0x02:
					t=ReadDate(offset);
					sendDates[i]=t;
					sendDate=true;
					break;
				case 0x0E:
					senderAddresses.Add(ReadString(offset));
					senderAddress=true;
					break;
				case 0x0D:
					senders.Add(ReadString(offset));
					sender=true;
					break;
				case 0x08:
					subjects.Add(ReadString(offset));
					subject=true;
					break;
				case 0x12:
					t=ReadDate(offset);
					receiveDates[i]=t;
					receiveDate=true;
					break;
				case 0x13:
					receivers.Add(ReadString(offset));
					receiver=true;
	    		break;
				case 0x14:
					receiverAddresses.Add(ReadString(offset));
					receiverAddress=true;
					break;
			}
			pos+=4;
		}

		if(!sender)
		{
			senders.Add(NULL);
		}
		if(!senderAddress)
		{
			senderAddresses.Add(NULL);
		}
		if(!receiver)
		{
			receivers.Add(NULL);
		}
		if(!receiverAddress)
		{
			receiverAddresses.Add(NULL);
		}
		if(!subject)
		{
			subjects.Add(NULL);
		}
		if(!receiveDate)
		{
			receiveDates[i]=NULL;
		}
		if(!sendDate)
		{
			sendDates[i]=NULL;
		}
	}
}

CString CDbxReader::ReadString(DWORD offset)
{
	if(offset>=f->GetLength())
	{
		AfxThrowFileException(CFileException::badSeek);
	}

	f->Seek(offset,CFile::begin);
	char c[256];
	c[255]='\0';
	CString s("");
	do
	{
		f->Read(c,255);
		s+=c;
	}
	while(strlen(c)==255);

	return s;
}

CTime CDbxReader::ReadDate(DWORD offset)
{
	if(offset>=f->GetLength())
	{
		AfxThrowFileException(CFileException::badSeek);
	}

	f->Seek(offset,CFile::begin);
	FILETIME fileTime;
	f->Read(&fileTime,8);
	CTime t(fileTime);
	if(t==-1) return CTime(1980,1,1,0,0,0);

	return t;
}

void CDbxReader::ReadIndexes()
{
	int INDEX_POINTER=0xE4;
	int ITEM_COUNT=0xC4;

	DWORD indexPtr=0;
	DWORD itemCount=0;
		
	f->Seek(INDEX_POINTER,CFile::begin);
	f->Read(&indexPtr,4);

	f->Seek(ITEM_COUNT,CFile::begin);
	f->Read(&itemCount,4);

	if(itemCount>0)
	{
		ReadIndex(indexPtr);
	}
}


void CDbxReader::ReadIndex(DWORD pos)
{
	if(pos>=f->GetLength())
	{
		AfxThrowFileException(CFileException::badSeek);
	}

	DWORD nextTable;
	BYTE ptrCount=0;
	DWORD indexCount;

	f->Seek(pos+8,CFile::begin);
	f->Read(&nextTable,4);
	f->Seek(5,CFile::current);
	f->Read(&ptrCount,1);
	f->Seek(2,CFile::current);
	f->Read(&indexCount,4);
	
	if(indexCount>0)
	{
		ReadIndex(nextTable);
	}

	pos+=24;
	
	int i;
	for(i=0;i<ptrCount;i++)
	{
		f->Seek(pos,CFile::begin);
		int indexPtr;
		f->Read(&indexPtr,4);
		f->Read(&nextTable,4);
		f->Read(&indexCount,4);
		
		indexes.Add(indexPtr);

		pos+=12;
		if(indexCount>0) ReadIndex(nextTable);
	}
}
